/*
 * AugmentedChartParser.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 14 May 2002
 * Time-stamp: <Thu May 22 20:25:51 EDT 2003 ferguson>
 */

/**
 * This is the augmented bottom-up parser described in NLU Chapter 4.
 * It uses features, but unless I get around to it, there's no
 * morphological analyzer, so you need to translate words like
 * ``sees'' into ``SEE +S'' and ``seeing'' into ``SEE +ING''.
 */
public class AugmentedChartParser extends BottomUpChartParser {
    /**
     * From BottomUpChartParser:
     * If the constituent of the given Entry matches the next Constituent
     * of the given Arc, then create and return a new Arc extending the given
     * one, otherwise return null.
     * <P>
     * The augmented parser differs from the basic one in that
     * (a) when extending an arc, the constituents are unified rather
     * than simply matching their categories, and (b) the rule used
     * to instantiate the new arc has the unifying substitution applied.
     * <P>
     * Implementation note: The sort of ugly casting here is to avoid
     * having to redo the entire ChartParser class using ``Augmented''
     * version of all the data structures (like Arc, Entry, etc.), which
     * are really the same except for using AugmentedConstituent and
     * AugmentedRule.
     */
    Arc extendArcWithEntry(Arc arc, Entry entry) {
	AugmentedConstituent constit = (AugmentedConstituent)entry.getConstituent();
	AugmentedConstituent nextconstit = (AugmentedConstituent)arc.getNextConstituent();
	Debug.debug("Attempting to extend " + arc);
	Debug.debug("  Next constit is: " + nextconstit);
	BindingSet bindings = new BindingSet();
	if (constit.unifyWith(nextconstit, bindings)) {
	    AugmentedGrammarRule rule = (AugmentedGrammarRule)arc.getRule();
	    return new Arc(rule.copyWithBindings(bindings),
			   arc.getStart(), entry.getEnd(),
			   arc.getNext() + 1);
	} else {
	    return null;
	}
    }
    /**
     * Test the augmented bottom-up chart parser.
     */
    public static void main(String argv[]) {
	main(argv, AugmentedLexicon.class, AugmentedGrammar.class, AugmentedChartParser.class);
    }
}
