/*
 * AugmentedGrammar.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 14 May 2002
 * Time-stamp: <Thu May 22 15:33:18 EDT 2003 ferguson>
 */

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * An AugmentedGrammar is one whose rules are AugmentedGrammarRules (ie,
 * whose rules have constituents that are AugmentedConstituents).
 * <P>
 * The main difference from a basic Grammar is that the rules returned
 * by getRulesForConstituent() need to unify their rhs with the given
 * constituent (rather than just matching syntactic category) and they
 * need to be copied and the unifying substitution applied.
 */
public class AugmentedGrammar extends Grammar {
    /**
     * Adds the given AugmentedGrammarRule to this AugmentedGrammar.
     * We override our parent's method for this operation in order to
     * ensure that only AugmentedGrammarRules are added to AugmentedGrammars.
     * We will throw a ClassCastException if the given object is not
     * a AugmentedGrammarRule. This is better than having it come up later
     * when we try to use the rule.
     */
    public void add(GrammarRule rule) {
	super.add((AugmentedGrammarRule)rule);
    }
    /**
     * Retrieve the List of GrammarRules whose first rhs constituent
     * unifies with the given AugmentedConstituent (and which have the
     * unifying subsitution applied).
     */
    // I would have thought I could use AugmentedConstituent as the parameter
    // here and it would automatically match, but the paren'ts method is
    // being called instead of this one, so we need to override more
    // obviously.
    public List getRulesForConstituent(Constituent constit) {
	AugmentedConstituent aconstit = (AugmentedConstituent)constit;
	Debug.debug("AugmentedGrammar.getRulesForConstituent: " + aconstit); 
	List result = new LinkedList();
	Iterator rules = super.getRulesForConstituent(aconstit).iterator();
	try {
	    while (rules.hasNext()) {
		AugmentedGrammarRule rule = (AugmentedGrammarRule)rules.next();
		AugmentedConstituent rhsconstit = (AugmentedConstituent)rule.getRhs().get(0);
		// Could just allocate this once and then clear each iteration
		BindingSet bindings = new BindingSet();
		if (rhsconstit.unifyWith(aconstit, bindings)) {
		    result.add(rule.copyWithBindings(bindings));
		}
	    }
	} catch (NoSuchElementException ex) {
	}
	Debug.debug("AugmentedGrammar.getRulesForConstituent: done"); 
	return result;
    }
    /**
     * Read the given filename and augment this AugmentedGrammar.
     * @see AugmentedGrammarFileHandler
     */
    public void readFile(String filename) throws IOException {
	new AugmentedGrammarFileHandler(this).readFile(filename);
    }
}
