/*
 * AugmentedGrammarFileHandler.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 10 May 2002
 * Time-stamp: <Fri May 23 11:22:03 EDT 2003 ferguson>
 */

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * An AugmentedGrammarFileHandler reads grammars whose constituents are
 * AugmentedConstituents.
 * <P>
 * The BNF grammar we are trying to parse is something like the
 * following:
 * <PRE>
    Grammar <- START("Grammar") Description? GrammarRule* END("Grammar")
    Description <- START(DESCRIPTION) characters END(Description)
    GrammarRule <- START("Rule") Constit+ END("Rule")
    Constit <- START(CATEGORY) Feature* END(CATEGORY)
    Feature <- START(FEATNAME) FeatureValue END(FEATNAME)
    FeatureValue <- characters
 * </PRE>
 * where CATEGORY is a syntactic category, and FEATNAME is a feature name.
 * These might be specified in the preamble of the file, or we could just
 * accept anything (which is in fact what we do now).
 */
public class AugmentedGrammarFileHandler extends GrammarFileHandler {
    /**
     * Create a new AugmentedGrammarFileHandler to read into the given
     * AugmentedGrammar.
     */
    public AugmentedGrammarFileHandler(AugmentedGrammar g) {
	super(g);
    }
    //
    // Finite-state machine parser starts here
    //
    static final int STATE_FEATURE = 5;
    Symbol feature;
    FeatureSet features;
    HashMap varTable;
    public void startElement(String namespaceURI, String sName, String qName,
                             Attributes attrs)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("augmented: START " + ename + ", state=" + state);
	switch (state) {
	case STATE_GRAMMAR:
	    if (ename.equalsIgnoreCase("rule")) {
		varTable = new HashMap();
	    }
	    break;
	case STATE_RULE:
	    features = new FeatureSet();
	    break;
	case STATE_CONSTIT:
	    // Starting a feature of the current constit
	    feature = Symbol.intern(ename);
	    state = STATE_FEATURE;
	    break;
	case STATE_FEATURE:
	    // Will be used for expressions nested in feature values
	    return;
	}
	// Unless consumed above...
	super.startElement(namespaceURI, sName, qName, attrs);
    }
    public void endElement(String namespaceURI, String sName, String qName)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("augmented: END " + ename);
	switch (state) {
	case STATE_RULE:
	    // Add rule to grammar
	    AugmentedConstituent lhs = (AugmentedConstituent)constits.removeFirst();
	    List rhs = constits;
	    grammar.add(new AugmentedGrammarRule(ruleId, lhs, rhs));
	    // Continue with grammar
	    state = STATE_GRAMMAR;
	    return;
	case STATE_CONSTIT:
	    // Make sure we're closing the constit we started
	    if (!ename.equalsIgnoreCase(cat.toString())) {
		ERROR("expected /" + cat + "; got /" + ename);
	    }
	    constits.add(new AugmentedConstituent(cat, features));
	    // Consit is finished: continue with rule
	    state = STATE_RULE;
	    return;
	case STATE_FEATURE:
	    // Make sure we're closing the feature we started
	    if (!ename.equalsIgnoreCase(feature.toString())) {
		ERROR("expected /" + feature + "; got /" + ename);
	    }
	    // Feature is finished: continue with constit
	    state = STATE_CONSTIT;
	    return;
	}
	// Unless consumed above...
	super.endElement(namespaceURI, sName, qName);
    }
    public void characters(char buf[], int offset, int len)
						throws SAXException
    {
	String text = new String(buf, offset, len).trim();
	if (text.equals("")) {
	    return;
	}
	Debug.debugIO("augmented: TEXT: \"" + text + "\"");
	switch (state) {
	case STATE_FEATURE:
	    Unifiable value = null;
	    if (text.startsWith("?")) {
		// Variable
		String name = null;
		VariableValueSet values = null;
		if (text.indexOf(' ') == -1) {
		    // Unconstrained variable: ?VARNAME
		    name = text.substring(1);
		} else {
		    // Constrained variable
		    StringTokenizer tokens = new StringTokenizer(text);
		    try {
			name = tokens.nextToken();
			if (name.equals("?")) {
			    // Spec is "? VARNAME ..."
			    name = tokens.nextToken();
			} else {
			    // Spec is "?VARNAME ..."
			    name = name.substring(1);
			}
			values = new VariableValueSet();
			while (tokens.hasMoreTokens()) {
			    String token = tokens.nextToken();
			    values.add(Symbol.intern(token));
			}
		    } catch (NoSuchElementException ex) {
			ERROR("bad variable spec");
		    }
		}
		// Check for var already in rule
		name = name.toUpperCase();
		value = (Variable)varTable.get(name);
		if (value == null) {
		    value = new Variable(Symbol.intern(name), values);
		    varTable.put(name, value);
		}
	    } else if (text.indexOf(' ') != -1) {
		ERROR("value is neither Variable nor Symbol: " + text);
	    } else {
		// Symbol
		value = Symbol.intern(text);
	    }
	    // Save this feature/value
	    features.setFeatureValue(feature, value);
	    return;
	}
	// Unless consumed above...
	super.characters(buf, offset, len);
    }
    //
    // Testing
    //
    public static void main(String argv[]) throws IOException {
	AugmentedGrammar grammar = new AugmentedGrammar();
	for (int i=0; i < argv.length; i++) {
	    grammar.readFile(argv[i]);
	}
	grammar.dump();
    }
}
