/*
 * AugmentedLexiconFileHandler.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 26 Apr 2002
 * Time-stamp: <Fri May 23 11:22:14 EDT 2003 ferguson>
 */

import java.io.File;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * An AugmentedLexiconFileHandler reads lexicon files whose constitents are
 * AugmentedConstituents.
 * <P>
 * The BNF grammar we are trying to parse is something like the
 * following:
 * <PRE>
    Lexicon <- START("Lexicon") Description? Word* END("Lexicon")
    Description <- START("Description") characters END("Description")
    Word <- START("Word" id=<word>) Sense+ END("Word")
    Sense <- START("Sense") Feature+ END("Sense")
    Feature <- START(FEATNAME) characters END(FEATNAME)
 * </PRE>
 * where FEATNAME is a feature name (such as "category", "root", etc.).
 * The value of the feature is parsed as a Symbol or Variable (possibly
 * constrained), although this syntax isn't described in XML.
 */
public class AugmentedLexiconFileHandler extends LexiconFileHandler {
    /**
     * Create a new AugmentedLexiconFileHandler to read into the given
     * AugmentedLexicon.
     */
    public AugmentedLexiconFileHandler(AugmentedLexicon l) {
	super(l);
    }
    //
    // Finite-state machine parser starts here
    //
    static final int STATE_FEATURE = 6;
    Symbol feature;
    FeatureSet features;
    public void startElement(String namespaceURI, String sName, String qName,
                             Attributes attrs)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("augmented: START " + ename + ", state=" + state);
	switch (state) {
	case STATE_WORD:
	    if (ename.equalsIgnoreCase("sense")) {
		features = new FeatureSet();
	    } else {
		ERROR("unexpected element in word: " + ename);
	    }
	    break;
	case STATE_SENSE:
	    if (!ename.equalsIgnoreCase("category")) {
		feature = Symbol.intern(ename);
		state = STATE_FEATURE;
		return;
	    }
	    break;
	case STATE_FEATURE:
	    // Until we have embedded elements in feature values...
	    ERROR("unexpected element in feature: " + ename);
	}
	// Unless consumed above...
	super.startElement(namespaceURI, sName, qName, attrs);
    }
    public void endElement(String namespaceURI, String sName, String qName)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("augmented: END " + ename);
	switch (state) {
	case STATE_SENSE:
	    if (ename.equalsIgnoreCase("sense")) {
		// Add entry to lexicon for this sense of word
		lexicon.add(word, new AugmentedConstituent(cat, features));
		// Sense is finished, continue with other senses of word
		state = STATE_WORD;
	    } else {
		ERROR("expected /sense; got /" + ename);
	    }
	    return;
	case STATE_FEATURE:
	    // We should be closing the feature we started
	    if (ename.equalsIgnoreCase(feature.toString())) {
		// Feature is finished: continue with this sense of word
		state = STATE_SENSE;
	    } else {
		ERROR("expected /" + feature + "; got /" + ename);
	    }
	    return;
	}
	// Unless consumed above...
	super.endElement(namespaceURI, sName, qName);
    }
    public void characters(char buf[], int offset, int len)
						throws SAXException
    {
	String text = new String(buf, offset, len).trim();
	if (text.equals("")) {
	    return;
	}
	Debug.debugIO("augmented: TEXT: \"" + text + "\"");
	switch (state) {
	case STATE_FEATURE:
	    Unifiable value = null;
	    if (text.startsWith("?")) {
		// Variable
		if (text.indexOf(' ') == -1) {
		    // Unconstrained variable: ?VARNAME
		    value = new Variable(Symbol.intern(text.substring(1)));
		} else {
		    // Constrained variable
		    String name = "";
		    StringTokenizer tokens = new StringTokenizer(text);
		    try {
			name = tokens.nextToken();
			if (name.equals("?")) {
			    // Spec is "? VARNAME ..."
			    name = tokens.nextToken();
			} else {
			    // Spec is "?VARNAME ..."
			    name = name.substring(1);
			}
			VariableValueSet values = new VariableValueSet();
			while (tokens.hasMoreTokens()) {
			    String token = tokens.nextToken();
			    values.add(Symbol.intern(token));
			}
			value = new Variable(Symbol.intern(name), values);
		    } catch (NoSuchElementException ex) {
			ERROR("bad variable spec");
		    }
		}
	    } else if (text.indexOf(' ') != -1) {
		ERROR("value is neither Variable nor Symbol: " + text);
	    } else {
		// Symbol
		value = Symbol.intern(text);
	    }
	    // Save this feature/value
	    features.setFeatureValue(feature, value);
	    return;
	}
	// Unless consumed above...
	super.characters(buf, offset, len);
    }
    //
    // Testing
    //
    public static void main(String argv[]) throws IOException {
	AugmentedLexicon lexicon = new AugmentedLexicon();
	for (int i=0; i < argv.length; i++) {
	    lexicon.readFile(argv[i]);
	}
	lexicon.dump();
    }
}
