/*
 * GapFeaturedGrammar.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 14 May 2002
 * Time-stamp: <Wed May 28 22:28:40 EDT 2003 ferguson>
 */

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * An GapFeaturedGrammar is one whose rules are HeadFeaturedGrammarRules
 * whose constituents can contain a GAP feature.
 * <P>
 * The difference between this and a HeadFeaturedGrammar is the special
 * expansion of rules that have a constituent with a GAP feature (and the
 * necessity to specify what constitute lexical categories).
 */
public class GapFeaturedGrammar extends HeadFeaturedGrammar {
    /**
     * The list of lexical categories (could be a set but...).
     */
    List lexicalCategories;
    /**
     * Sets the list of lexical categories for this GapFeaturedGrammar.
     */
    public void setLexicalCategories(List cats) {
	this.lexicalCategories = cats;
    }
    /**
     * Returns the list of features that are head features for the
     * given syntactic category (or null if none are defined).
     */
    public boolean isLexicalCategory(Symbol category) {
	return (lexicalCategories != null &&
		lexicalCategories.contains(category));
    }
    /**
     * Adds the given HeadFeaturedGrammarRule to this GapFeaturedGrammar.
     * For the GapFeaturedGrammar, we have to expand the rule using
     * the algorithm for handling GAP features (Figure 5.5 in NLU2), in
     * addition to the normal processing of head features for a
     * HeadFeaturedGrammar.
     */
    public void add(GrammarRule rule) {
	HeadFeaturedGrammarRule augrule = (HeadFeaturedGrammarRule)rule;
	Iterator newrules = expandGapFeatures(augrule).iterator();
	try {
	    while (newrules.hasNext()) {
		augrule = (HeadFeaturedGrammarRule)newrules.next();
		System.err.println("adding rule: " + augrule);
		super.add(augrule);
	    }
	} catch (NoSuchElementException ex) {
	}
    }
    /**
     * Returns a list of rules resulting from expanding the the GAP
     * features of the given HeadFeaturedGrammarRule. The algorithm
     * (Figure 5.5 in NLU2) is:
     * <p>
     * For each rule Y -> X1 ... Hi ... Xn with head constituent Hi
     * <ol>
     * <li>If the rule specifies a GAP feature in some constituent
     *     already, then skip.
     * <li>If the head Hi is not a lexical category, then add a GAP
     *     feature to the head and the mother, and (GAP -) to the other
     *     subconstituents, producing a rule of the form:
     * <pre>
     * (Y GAP ?g) -> (X1 GAP -) ... (Hi GAP ?g) ... (Xn GAP -)
     * </pre>
     * <li>If the head Hi is a lexical category, then for each non-lexical
     *     subconstituent Xj, add a rule of the form:
     * <pre>
     * (Y GAP ?g) -> (X1 GAP -) ... (Xj GAP ?g) ... (Xn GAP -)
     * </pre>
     * </ol>
     */
    List expandGapFeatures(HeadFeaturedGrammarRule rule) {
	System.err.println("expandGapFeatures: " + rule);
	// First see if the rule has a GAP anywhere
	if (hasGap(rule)) {
	    // Yes => don't expand it
	    System.err.println("already has GAP; not expanding");
	} else {
	    HeadFeaturedConstituent head = rule.getHeadConstituent();
	    System.err.println("head=" + head);
	    if (head == null) {
		Debug.warn("no head in rule: " + rule);
	    } else {
		Symbol headCat = head.getCategory();
		System.err.println("headCat=" + headCat);
		// If head constituent specifies a lexical category...
		if (!isLexicalCategory(headCat)) {
		    // ...Then gap is for it and mother only
		    System.err.println("head is not lexical");
		    addGapForHead(rule);
		} else {
		    // Otherwise gap is for all non-lexical constits
		    System.err.println("head is lexical");
		    return addGapsForNonLexicalConstits(rule);
		}
	    }
	}
	// Unless something else was returned above, return just this rule
	// (which may have been modified)
	List result = new LinkedList();
	result.add(rule);
	return result;
    }
    /**
     * Returns true if any constituent of the given rule specifies
     * a GAP feature.
     */
    boolean hasGap(HeadFeaturedGrammarRule rule) {
	if (hasGap((AugmentedConstituent)rule.getLhs())) {
	    return true;
	} else {
	    Iterator constits = rule.getRhs().iterator();
	    try {
		while (constits.hasNext()) {
		    AugmentedConstituent constit = (AugmentedConstituent)constits.next();
		    if (hasGap(constit)) {
			return true;
		    }
		}
	    } catch (NoSuchElementException ex) {
	    }
	    return false;
	}
    }	
    /**
     * Returns true if the given constituent specifies a GAP feature.
     */
    boolean hasGap(AugmentedConstituent constit) {
	return (constit.getFeatures().getFeatureValue(SYMBOL_GAP) != null);
    }
    /**
     * The Symbol ``GAP'', used for indicating the GAP feature.
     */
    final Symbol SYMBOL_GAP = Symbol.intern("GAP");
    /**
     * The Symbol ``+'', used for indicating the GAP feature's value.
     */
    final Symbol SYMBOL_PLUS = Symbol.intern("+");
    /**
     * Step 2 of the GAP feature algorithm (Figure 5.5 in NLU2).
     * If the head Hi is not a lexical category, then add a GAP
     * feature to the head and the mother, and (GAP -) to the other
     * subconstituents, producing a rule of the form:
     * <pre>
     * (Y GAP ?g) -> (X1 GAP -) ... (Hi GAP ?g) ... (Xn GAP -)
     * </pre>
     * Note that (GAP -) is implied, so we don't have to add it explicitly.
     */
    void addGapForHead(HeadFeaturedGrammarRule rule) {
	System.err.println("adding gap for head: " + rule);
	AugmentedConstituent mother = (AugmentedConstituent)rule.getLhs();
	AugmentedConstituent head = (AugmentedConstituent)rule.getHeadConstituent();
	Variable var = new Variable(Symbol.gensym("G"));
	mother.getFeatures().setFeatureValue(SYMBOL_GAP, var);
	head.getFeatures().setFeatureValue(SYMBOL_GAP, var);
    }
    /**
     * Step 3 of the GAP feature algorithm (Figure 5.5 in NLU2).
     * <li>If the head Hi is a lexical category, then for each non-lexical
     *     subconstituent Xj, add a rule of the form:
     * <pre>
     * (Y GAP ?g) -> (X1 GAP -) ... (Xj GAP ?g) ... (Xn GAP -)
     * </pre>
     */
    List addGapsForNonLexicalConstits(HeadFeaturedGrammarRule rule) {
	System.err.println("gapping non-lexical children: " + rule);
	List result = new LinkedList();
	List rhs = rule.getRhs();
	for (int i=0; i < rhs.size(); i++) {
	    Constituent constit = (Constituent)rhs.get(i);
	    System.err.println("constit: " + constit);
	    if (!isLexicalCategory(constit.getCategory())) {
		System.err.println("constit is non-lexical");
		result.add(makeGapRule(rule, i));
	    }
	}
	return result;
    }
    /**
     * Returns a copy of the given rule but with a GAP feature added
     * connecting the mother and the i'th child.
     */
    HeadFeaturedGrammarRule makeGapRule(HeadFeaturedGrammarRule rule, int i) {
	/* TOODODODODO
	HeadFeaturedGrammarRule newrule = rule.clone();
	AugmentedConstituent mother = (AugmentedConstituent)newrule.getLhs();
	AugmentedConstituent child = (AugmentedConstituent)newrule.getRhs().get(i);
	Variable var = new Variable(Symbol.gensym("G"));
	mother.getFeatures().setFeatureValue(SYMBOL_GAP, var);
	child.getFeatures().setFeatureValue(SYMBOL_GAP, var);
	return newrule;
	*/
	return rule;
    }
    /**
     * Read the given filename and augment this GapFeaturedGrammar.
     * @see GapFeaturedGrammarFileHandler
     */
    public void readFile(String filename) throws IOException {
	new GapFeaturedGrammarFileHandler(this).readFile(filename);
    }
}
