/*
 * GapFeaturedGrammarFileHandler.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 10 May 2002
 * Time-stamp: <Wed May 28 21:55:16 EDT 2003 ferguson>
 *
 */

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * An GapFeaturedGrammarFileHandler reads grammars whose constituents are
 * HeadFeaturedConstituents.
 * The main difference from the HeadFeaturedGrammarFileHandler is that
 * we need to properly parse embedded constituents as the value of
 * the GAP feature.
 * @see GapFeaturedChartParser
 */
public class GapFeaturedGrammarFileHandler extends HeadFeaturedGrammarFileHandler {
    /**
     * Create a new GapFeaturedGrammarFileHandler to read into the given
     * GapFeaturedGrammar.
     */
    public GapFeaturedGrammarFileHandler(GapFeaturedGrammar g) {
	super(g);
    }
    //
    // Finite-state machine parser starts here
    //
    static final int STATE_LEXICALCATEGORIES = 7;
    boolean isEmbedded = false;
    Symbol savedCat;
    Symbol savedFeature;
    FeatureSet savedFeatures;
    boolean savedIsHead;
    public void startElement(String namespaceURI, String sName, String qName,
                             Attributes attrs)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("gap: START " + ename + ", state=" + state);
	switch (state) {
	case STATE_GRAMMAR:
	    // Look for lexicalcategories element
	    if (ename.equalsIgnoreCase("lexicalcategories")) {
		state = STATE_LEXICALCATEGORIES;
	    } else {
		super.startElement(namespaceURI, sName, qName, attrs);
	    }
	    return;
	case STATE_FEATURE:
	    // Begin embedded constituent (see AugmentedGrammarFileHandler)
	    Debug.debugIO("begin embedded constit");
	    isEmbedded = true;
	    // Save state of parse of parent constit
	    savedCat = cat;
	    savedFeature = feature;
	    savedFeatures = features;
	    savedIsHead = isHead;
	    state = STATE_RULE;
	    // Pass on to parent to process start of constit
	    break;
	}
	// Unless handled above...
	super.startElement(namespaceURI, sName, qName, attrs);
    }
    public void endElement(String namespaceURI, String sName, String qName)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("gap: END " + ename + ", state = " + state + ", cat = " + cat);
	switch (state) {
	case STATE_CONSTIT:
	    // Make sure we're closing the constit we started
	    if (!ename.equalsIgnoreCase(cat.toString())) {
		ERROR("expected /" + cat + "; got /" + ename);
	    }
	    // If this constituent was embedded...
	    if (isEmbedded) {
		Debug.debugIO("end embedded constit");
		// ...Then we use it as the value of the current feature
		savedFeatures.setFeatureValue(savedFeature, new AugmentedConstituent(cat, features));
		// Restore state of parent parse
		isEmbedded = false;
		cat = savedCat;
		feature = savedFeature;
		features = savedFeatures;
		isHead = savedIsHead;
		state = STATE_FEATURE;
		return;
	    }
	    break;
	case STATE_LEXICALCATEGORIES:
	    if (ename.equalsIgnoreCase("LEXICALCATEGORIES")) {
		// End of specifying lexical categories: rules should be next
		state = STATE_GRAMMAR;
	    } else {
		ERROR("expected /lexicalcategories; got /" + ename);
	    }
	    return;
	}
	// Unless handled above...
	super.endElement(namespaceURI, sName, qName);
    }
    public void characters(char buf[], int offset, int len)
						throws SAXException
    {
	String text = new String(buf, offset, len).trim();
	if (text.equals("")) {
	    return;
	}
	Debug.debugIO("gap: TEXT: \"" + text + "\"");
	switch (state) {
	case STATE_LEXICALCATEGORIES:
	    // List of categories considered ``lexical''
	    List names = new LinkedList();
	    StringTokenizer tokens = new StringTokenizer(text);
	    try {
		while (tokens.hasMoreTokens()) {
		    String token = tokens.nextToken();
		    names.add(Symbol.intern(token));
		}
	    } catch (NoSuchElementException ex) {
	    }
	    ((GapFeaturedGrammar)grammar).setLexicalCategories(names);
	    return;
	}
	// Unless handled above...
	super.characters(buf, offset, len);
    }

    public static void main(String argv[]) throws IOException {
	//Debug.setDebugIO(true);
	GapFeaturedGrammar grammar = new GapFeaturedGrammar();
	for (int i=0; i < argv.length; i++) {
	    grammar.readFile(argv[i]);
	}
	grammar.dump();
    }
}
