/*
 * HeadFeaturedGrammar.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 14 May 2002
 * Time-stamp: <Thu May 22 15:46:45 EDT 2003 ferguson>
 */

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * An HeadFeaturedGrammar is one whose rules are HeadFeaturedGrammarRules (ie,
 * whose rules have constituents that are HeadFeaturedConstituents, ie.,
 * they can be marked as head constituents).
 * <P>
 * The only difference from an AugmentedGrammar is that (a) the grammar
 * can specify the head features for syntactic categories, and (b) the
 * HeadFeaturedGrammarRules will have these head features automatically
 * added (and properly connected to the head daughter constituent).
 */
public class HeadFeaturedGrammar extends AugmentedGrammar {
    /**
     * The set of head features defined for various categories.
     */
    AssociationList headFeatures = new AssociationList();
    /**
     * Sets the list of features that are head features for the
     * given syntactic category.
     */
    public void setHeadFeatures(Symbol category, List names) {
	headFeatures.put(category, names);
    }
    /**
     * Returns the list of features that are head features for the
     * given syntactic category (or null if none are defined).
     */
    public List getHeadFeatures(Symbol category) {
	return (List)headFeatures.get(category);
    }
    /**
     * Adds the given HeadFeaturedGrammarRule to this HeadFeaturedGrammar.
     * As with AugmentedGrammarRule, we override our parent's method for this operation in order to
     * ensure that only HeadFeaturedGrammarRules are added to HeadFeaturedGrammars.
     * We will throw a ClassCastException if the given object is not
     * a HeadFeaturedGrammarRule. This is better than having it come up later
     * when we try to use the rule.
     * <P>
     * Also, for the HeadFeaturedGrammar, we have to add the head features from
     * the head constituent of the lhs of the rule to the rhs. After this has
     * been done, no further specialization is needed for the AugmentedChartParser
     * code to work properly.
     */
    public void add(GrammarRule rule) {
	HeadFeaturedGrammarRule augrule = (HeadFeaturedGrammarRule)rule;
	addHeadFeatures(augrule);
	super.add(augrule);
    }
    /**
     * Adds the features from the head daughter of the given rule to its
     * mother.
     * We must check both the mother and head to see if these features
     * already are defined.
     * Note that the features are inserted into both the mother and the
     * head daughter if they're not already there.
     */
    void addHeadFeatures(HeadFeaturedGrammarRule rule) {
	//System.err.println("addHeadFeatures: " + rule);
	HeadFeaturedConstituent mother = (HeadFeaturedConstituent)rule.getLhs();
	//System.err.println("mother=" + mother);
	HeadFeaturedConstituent head = rule.getHeadConstituent();
	//System.err.println("head=" + head);
	if (head == null) {
	    Debug.warn("no head in rule: " + rule);
	} else {
	    // Relevant head features are those for the mother
	    Symbol motherCat = mother.getCategory();
	    //System.err.println("motherCat=" + motherCat);
	    List motherCatHeadFeatures = getHeadFeatures(motherCat);
	    if (motherCatHeadFeatures != null) {
		FeatureSet motherFeatures = mother.getFeatures();
		//System.err.println("motherFeatures=" + motherFeatures);
		FeatureSet headFeatures = head.getFeatures();
		//System.err.println("headFeatures=" + headFeatures);
		Iterator categoryHeadFeatures = motherCatHeadFeatures.iterator();
		try {
		    while (categoryHeadFeatures.hasNext()) {
			Symbol featname = (Symbol)categoryHeadFeatures.next();
			//System.err.println("featname=" + featname);
			Unifiable mval = motherFeatures.getFeatureValue(featname);
			//System.err.println("mval=" + mval);
			Unifiable hval = headFeatures.getFeatureValue(featname);
			//System.err.println("hval=" + hval);
			Symbol varname = Symbol.gensym(featname);
			Unifiable newval = null;
			if (mval == null && hval == null) {
			    // New unconstrained var for head feature
			    newval = new Variable(varname);
			} else if (mval == null) {
			    if (hval instanceof Variable) {
				// Use hval as var
				newval = hval;
			    } else {
				// New var constrained to value of hval
				VariableValueSet values = new VariableValueSet();
				values.add(hval);
				newval = new Variable(varname, values);
			    }
			} else if (hval == null || mval.equals(hval)) {
			    if (mval instanceof Variable) {
				// Use mval as var
				newval = mval;
			    } else {
				// New var constrained to value of mval
				VariableValueSet values = new VariableValueSet();
				values.add(mval);
				newval = new Variable(varname, values);
			    }
			} else {
			    Debug.warn("incompatible head feature: " + featname +
				       " in rule " + rule);
			}
			if (newval != null) {
			    //System.err.println("newval=" + newval);
			    if (mval == null) {
				//System.err.println("adding to mother");
				motherFeatures.setFeatureValue(featname, newval);
			    }
			    if (hval == null) {
				//System.err.println("adding to head daughter");
				headFeatures.setFeatureValue(featname, newval);
			    }
			}
		    }
		} catch (NoSuchElementException ex) {
		}
	    }
	    //System.err.println("done");
	}
    }
    /**
     * Read the given filename and augment this HeadFeaturedGrammar.
     * @see HeadFeaturedGrammarFileHandler
     */
    public void readFile(String filename) throws IOException {
	new HeadFeaturedGrammarFileHandler(this).readFile(filename);
    }
}
