/*
 * GrammarFileHandler.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 10 May 2002
 * Time-stamp: <Fri May 23 11:21:38 EDT 2003 ferguson>
 */

import java.io.FileReader;
import java.io.IOException;
import java.util.List;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * GrammarFileHandler reads the basic grammar files, whose constituents
 * are simple Constituents (no features).
 * <P>
 * The BNF grammar we are trying to parse is something like the
 * following:
 * <PRE>
    Grammar <- START("Grammar") Description? GrammarRule* END("Grammar")
    Description <- START(DESCRIPTION) characters END(Description)
    GrammarRule <- START("Rule") Constit+ END("Rule")
    Constit <- START(CAT) END(CAT)
 * </PRE>
 * where CAT is a syntactic category.
 * <P>
 * See implementation notes for LexiconFileHandler.
 */
public class GrammarFileHandler extends DefaultHandler {
    /**
     * The Grammar we are building while reading the file.
     */
    Grammar grammar;
    /**
     * Create a new GrammarFileHandler to read into the given Grammar.
     */
    public GrammarFileHandler(Grammar g) {
	super();
	this.grammar = g;
    }
    /**
     * Read the given filename using this handler.
     */
    public void readFile(String filename) throws IOException {
	try {
	    // Create XMLReader
	    XMLReader reader = XMLReaderFactory.createXMLReader();
	    reader.setContentHandler(this);
	    reader.setErrorHandler(this);
	    // To enforce the DTD...
	    try {
		reader.setFeature("http://xml.org/sax/features/validation", true);
	    } catch (SAXException e) {
		Debug.warn("can't enable validation for XMLReader");
	    }
	    // Parse the input
	    reader.parse(new InputSource(filename));
	} catch (SAXParseException ex) {
	    throw new IOException(ex.getSystemId() + ", line " +
				  ex.getLineNumber() + ": " +
				  ex.getMessage());
	} catch (SAXException ex) {
	    throw new IOException(ex.getMessage());
	}
    }
    //
    // Finite-state machine parser starts here
    //
    int state;
    static final int STATE_FILE = 0;
    static final int STATE_GRAMMAR = 1;
    static final int STATE_DESCRIPTION = 2;
    static final int STATE_RULE = 3;
    static final int STATE_CONSTIT = 4;
    StringBuffer textBuffer = new StringBuffer();
    Symbol ruleId;
    LinkedList constits;
    Symbol cat;
    /**
     * SAX ContentHandler method called when a start-element tag is
     * encountered.
     */
    public void startElement(String namespaceURI, String sName, String qName,
                             Attributes attrs)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("START " + ename + ", state=" + state);
	String id;
	switch (state) {
	case STATE_FILE:
	    // Can only start a grammar in this (initial) state
	    if (ename.equalsIgnoreCase("grammar")) {
		state = STATE_GRAMMAR;
	    } else {
		ERROR("unexpected element in file: " + ename);
	    }
	    break;
	case STATE_GRAMMAR:
	    if (ename.equalsIgnoreCase("description")) {
		state = STATE_DESCRIPTION;
	    } else if (ename.equalsIgnoreCase("rule")) {
		// Rule name comes as attribute
		id = XMLUtil.getAttributeValue(attrs, "id");
		ruleId = (id != null ? Symbol.intern(id) : Symbol.gensym("RULE"));
		constits = new LinkedList();
		state = STATE_RULE;
	    } else {
		ERROR("unexpected element in grammar: " + ename);
	    }
	    break;
	case STATE_RULE:
	    cat = Symbol.intern(ename);
	    state = STATE_CONSTIT;
	    break;
	case STATE_CONSTIT:
	    ERROR("unexpected element in constit: " + ename);
	    break;
	}
    }
    /**
     * SAX ContentHandler method called when an end-element tag is
     * encountered.
     */
    public void endElement(String namespaceURI, String sName, String qName)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("END " + ename);
	switch (state) {
	case STATE_FILE:
	    // Should never have an end tag in this state
	    ERROR("unexpected /" + ename);
	case STATE_GRAMMAR:
	    // Should only close grammar in this state
	    if (ename.equalsIgnoreCase("grammar")) {
		// Grammar is finished: I suppose we could start another one...
		state = STATE_FILE;
	    } else {
		ERROR("expected /grammar; got /" + ename);
	    }
	    break;
	case STATE_DESCRIPTION:
	    if (ename.equalsIgnoreCase("description")) {
		// End of description: look for rules
		grammar.setDescription(consumeText());
		state = STATE_GRAMMAR;
	    } else {
		ERROR("expected /description; got /" + ename);
	    }
	    break;
	case STATE_RULE:
	    if (ename.equalsIgnoreCase("rule")) {
		// Add rule to grammar
		Constituent lhs = (Constituent)constits.removeFirst();
		List rhs = constits;
		grammar.add(new GrammarRule(ruleId, lhs, rhs));
		// Continue with grammar
		state = STATE_GRAMMAR;
	    } else {
		ERROR("expected /rule; got /" + ename);
	    }
	    break;
	case STATE_CONSTIT:
	    // Make sure we're closing the constit we started
	    if (ename.equalsIgnoreCase(cat.toString())) {
		constits.add(new Constituent(cat));
		// Consit is finished: continue with rule
		state = STATE_RULE;
	    } else {
		ERROR("expected /" + cat + "; got /" + ename);
	    }
	    break;
	}
    }
    /**
     * Called by the SAX parsing engine when characters are read outside
     * of any element tag. This can be called multiple times, such as
     * when the text spans multiple lines (depends on whitespace treatment,
     * whether the parser is validating, etc.). So the safe thing to do is
     * just hold onto the text for later use when an end-element tag is
     * reached (assuming we don't care about the actual form of the
     * original text, but just its content, which is the case for us).
     * We could be more careful, saving the text in states where we're
     * expecting text and otherwise indicating a state-specific error,
     * but frankly I don't care that much.
     */
    public void characters(char buf[], int offset, int len)
						throws SAXException
    {
	String text = new String(buf, offset, len).trim();
	if (text.equals("")) {
	    return;
	}
	Debug.debugIO("TEXT: \"" + text + "\"");
	if (textBuffer.length() > 0) {
	    // I suppose appending a space is the right thing to do...
	    textBuffer.append(" ");
	}
	textBuffer.append(text);
    }
    /**
     * Returns the contents of the textBuffer and clears the buffer.
     */
    String consumeText() {
	String text = textBuffer.toString();
	textBuffer.setLength(0);
	return text;
    }
    /**
     * Locator used in producing error messages.
     */
    Locator locator;
    /**
     * SAX ContentHandler method called at the start of processing a
     * document.
     */
    public void setDocumentLocator(Locator locator) {
	this.locator = locator;
    }
    /**
     * Generates a SAXParseException with the given message and the
     * current locator.
     */
    void ERROR(String msg) throws SAXParseException {
	throw new SAXParseException(msg, locator);
    }
    //
    // ErrorHandler methods
    //
    /**
     * Receive notification of a recoverable error.
     */
    public void error(SAXParseException ex) {
	Debug.warn(ex.getSystemId() + ", line " +
		   ex.getLineNumber() + ": " +
		   ex.getMessage());
    }
    /**
     * Receive notification of a non-recoverable error.
     */
    public void fatalError(SAXParseException ex) {
	Debug.warn(ex.getSystemId() + ", line " +
		   ex.getLineNumber() + ": " +
		   ex.getMessage());
    }
    /**
     * Receive notification of a warning.
     */
    public void warning(SAXParseException ex) {
	Debug.warn(ex.getSystemId() + ", line " +
		   ex.getLineNumber() + ": " +
		   ex.getMessage());
    }
    //
    // Testing
    //
    /**
     * Test the GrammarFileHandler.
     */
    public static void main(String argv[]) throws IOException {
	Grammar grammar = new Grammar();
	new GrammarFileHandler(grammar).readFile(argv[0]);
	grammar.dump();
    }
}
