/*
 * LexiconFileHandler.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 26 Apr 2002
 * Time-stamp: <Fri May 23 11:21:52 EDT 2003 ferguson>
 */

import java.io.File;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * LexiconFileHandler reads lexicon files whose constitents are simple
 * Constituents (no features).
 * <P>
 * The BNF grammar we are trying to parse is something like the
 * following:
 * <PRE>
    Lexicon <- START("Lexicon") Description? Word* END("Lexicon")
    Description <- START("Description") characters END("Description")
    Word <- START("Word" id=<word>) Sense+ END("Word")
    Sense <- START("Sense") Category END("Sense")
    Category <- START("Category") CATEGORY END("Category")
 * </PRE>
 * where CATEGORY is a syntactic category ("N", "V", etc.).
 * <P>
 * Implementation note: The SAX ContentHandler interface calls us when
 * start- and end-element tags are read, and for the text between them.
 * Using these callbacks, we need to maintain the state of our parse of
 * the file, and use that state to decide what to do with a particular
 * tag's start and end. Usually, handling a start-element tag involves
 * extracting attributes needed for the element, then change to state of
 * the parse to reflect that we are now processing that element. Then,
 * the handling of the matching end-element tag performs the operation
 * appropriate to the end of the element (usually creating some object and
 * saving it in a parent data structure), then setting the state back to
 * continue the parse.
 * <P>
 * It would be much more elegant to have a tool like yacc/bison which
 * compiled CFG specifications into Java code that used the SAX parser
 * as its tokenizer (input), and looked after all the state management.
 * Maybe some day...
 */
public class LexiconFileHandler extends DefaultHandler {
    /**
     * The Lexicon we are building while reading the file.
     */
    Lexicon lexicon;
    /**
     * Create a new LexiconFileHandler to read into the given Lexicon.
     * @see Lexicon#readFile
     */
    public LexiconFileHandler(Lexicon l) {
	super();
	this.lexicon = l;
    }
    /**
     * Read the given filename using this handler.
     */
    public void readFile(String filename) throws IOException {
	try {
	    // Create XMLReader
	    XMLReader reader = XMLReaderFactory.createXMLReader();
	    reader.setContentHandler(this);
	    reader.setErrorHandler(this);
	    // To enforce the DTD...
	    try {
		reader.setFeature("http://xml.org/sax/features/validation", true);
	    } catch (SAXException e) {
		Debug.warn("can't enable validation for XMLReader");
	    }
	    // Parse the input
	    reader.parse(new InputSource(filename));
	} catch (SAXParseException ex) {
	    throw new IOException(ex.getSystemId() + ", line " +
				  ex.getLineNumber() + ": " +
				  ex.getMessage());
	} catch (SAXException ex) {
	    throw new IOException(ex.getMessage());
	}
    }
    //
    // Finite-state machine parser starts here
    //
    int state;
    static final int STATE_FILE = 0;
    static final int STATE_LEXICON = 1;
    static final int STATE_DESCRIPTION = 2;
    static final int STATE_WORD = 3;
    static final int STATE_SENSE = 4;
    static final int STATE_CATEGORY = 5;
    StringBuffer textBuffer = new StringBuffer();
    String description;
    Symbol word;
    Symbol cat;
    /**
     * SAX ContentHandler method called when a start-element tag is
     * encountered.
     */
    public void startElement(String namespaceURI, String sName, String qName,
                             Attributes attrs)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("START " + ename + ", state=" + state);
	switch (state) {
	case STATE_FILE:
	    // Can only start a lexicon in this (initial) state
	    if (ename.equalsIgnoreCase("lexicon")) {
		state = STATE_LEXICON;
	    } else {
		ERROR("expected element in file: " + ename);
	    }
	    break;
	case STATE_LEXICON:
	    if (ename.equalsIgnoreCase("description")) {
		state = STATE_DESCRIPTION;
	    } else if (ename.equalsIgnoreCase("word")) {
		// Word itself comes as id attribute
		String id = XMLUtil.getAttributeValue(attrs, "id");
		if (id == null) {
		    ERROR("missing id attribute for word");
		} else {
		    word = Symbol.intern(id);
		}
		// Proceed with senses of word
		state = STATE_WORD;
	    } else {
		ERROR("unexpected element in lexicon: " + ename);
	    }
	    break;
	case STATE_DESCRIPTION:
	    ERROR("unexpected element in description: " + ename);
	case STATE_WORD:
	    if (ename.equalsIgnoreCase("sense")) {
		state = STATE_SENSE;
	    } else {
		ERROR("unexpected element in word: " + ename);
	    }
	    break;
	case STATE_SENSE:
	    if (ename.equalsIgnoreCase("category")) {
		state = STATE_CATEGORY;
	    } else {
		ERROR("unexpected element in sense: " + ename);
	    }
	    break;
	case STATE_CATEGORY:
	    ERROR("unexpected element in category: " + ename);
	}
    }
    /**
     * SAX ContentHandler method called when an end-element tag is
     * encountered.
     */
    public void endElement(String namespaceURI, String sName, String qName)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("END " + ename);
	switch (state) {
	case STATE_FILE:
	    // Should never have an end tag in this state
	    ERROR("unexpected /" + ename);
	case STATE_LEXICON:
	    // Should only close lexicon in this state
	    if (ename.equalsIgnoreCase("lexicon")) {
		// Lexicon is finished: I suppose we could start another one...
		state = STATE_FILE;
	    } else {
		ERROR("expected /lexicon; got /" + ename);
	    }
	    break;
	case STATE_DESCRIPTION:
	    // Set description of lexicon
	    lexicon.setDescription(consumeText());
	    // End of description: look for entries
	    state = STATE_LEXICON;
	    break;
	case STATE_WORD:
	    // Make sure we're closing the word (cat) we started
	    if (ename.equalsIgnoreCase("word")) {
		// Word is finished: continue with lexicon
		state = STATE_LEXICON;
	    } else {
		ERROR("expected /word; got /" + ename);
	    }
	    break;
	case STATE_SENSE:
	    if (ename.equalsIgnoreCase("sense")) {
		// Add entry to lexicon for this sense of word
		lexicon.add(word, new Constituent(cat));
		// Sense is finished, continue with other senses of word
		state = STATE_WORD;
	    } else {
		ERROR("expected /sense; got /" + ename);
	    }
	    break;
	case STATE_CATEGORY:
	    if (ename.equalsIgnoreCase("category")) {
		// Save category
		cat = Symbol.intern(consumeText());
		// Continue with this sense of word
		state = STATE_SENSE;
	    } else {
		ERROR("expected /category; got /" + ename);
	    }
	    break;
	}
    }
    /**
     * Called by the SAX parsing engine when characters are read outside
     * of any element tag. This can be called multiple times, such as
     * when the text spans multiple lines (depends on whitespace treatment,
     * whether the parser is validating, etc.). So the safe thing to do is
     * just hold onto the text for later use when an end-element tag is
     * reached (assuming we don't care about the actual form of the
     * original text, but just its content, which is the case for us).
     * We could be more careful, saving the text in states where we're
     * expecting text and otherwise indicating a state-specific error,
     * but frankly I don't care that much.
     */
    public void characters(char buf[], int offset, int len)
						throws SAXException
    {
	String text = new String(buf, offset, len).trim();
	if (text.equals("")) {
	    return;
	}
	Debug.debugIO("TEXT: \"" + text + "\"");
	if (textBuffer.length() > 0) {
	    // I suppose appending a space is the right thing to do...
	    textBuffer.append(" ");
	}
	textBuffer.append(text);
    }
    /**
     * Returns the contents of the textBuffer and clears the buffer.
     */
    String consumeText() {
	String text = textBuffer.toString();
	textBuffer.setLength(0);
	return text;
    }
    /**
     * Locator used in producing error messages.
     */
    Locator locator;
    /**
     * SAX ContentHandler method called at the start of processing a
     * document.
     */
    public void setDocumentLocator(Locator locator) {
	this.locator = locator;
    }
    /**
     * Generates a SAXParseException with the given message and the
     * current locator.
     */
    void ERROR(String msg) throws SAXParseException {
	throw new SAXParseException(msg, locator);
    }
    //
    // ErrorHandler methods
    //
    /**
     * Receive notification of a recoverable error.
     */
    public void error(SAXParseException ex) {
	Debug.warn(ex.getSystemId() + ", line " +
		   ex.getLineNumber() + ": " +
		   ex.getMessage());
    }
    /**
     * Receive notification of a non-recoverable error.
     */
    public void fatalError(SAXParseException ex) {
	Debug.warn(ex.getSystemId() + ", line " +
		   ex.getLineNumber() + ": " +
		   ex.getMessage());
    }
    /**
     * Receive notification of a warning.
     */
    public void warning(SAXParseException ex) {
	Debug.warn(ex.getSystemId() + ", line " +
		   ex.getLineNumber() + ": " +
		   ex.getMessage());
    }
    //
    // Testing
    //
    /**
     * Test the LexiconFileHandler.
     */
    public static void main(String argv[]) throws IOException {
	Lexicon lexicon = new Lexicon();
	new LexiconFileHandler(lexicon).readFile(argv[0]);
	lexicon.dump();
    }
}
