/*
 * Symbol.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 24 Apr 2002
 * Time-stamp: <Wed May 14 12:22:32 EDT 2003 ferguson>
 */

import java.util.HashMap;

/**
 * Symbols are a basic datatype in Lisp. There should be only one Symbol
 * with a given name accessible, so most applications should use
 * intern() rather than creating Symbols using the constructor.
 */
public class Symbol extends Object implements Unifiable {
    /**
     * The name of this symbol (what Lisp calls the ``print-name'').
     */
    String name;
    /**
     * Create a new symbol with the given name. You almost certainly want
     * use intern() rather than calling this constructor directly.
     */
    public Symbol(String name) {
	this.name = name;
    }
    /**
     * Returns the printable representation of this Symbol (i.e., its name).
     */
    public String toString() {
	return name;
    }
    /**
     * Private static (class) field holding name->Symbol mapping.
     */
    static HashMap symbolTable = new HashMap();
    /**
     * Main method for getting a Symbol given a name; creates the Symbol
     * if necessary. This method could be extended to support Lisp's
     * notion of ``packages'' if anyone cared. Note that Symbols are
     * not case-sensitive (on purpose).
     */
    public static Symbol intern(String name) {
	// Upcase string before checking symbol table
	name = name.toUpperCase();
	Symbol sym = (Symbol)symbolTable.get(name);
	if (sym == null) {
	    sym = new Symbol(name);
	    symbolTable.put(name, sym);
	}
	return sym;
    }
    /**
     * Return a new uniquely-named Symbol whose name starts with the given
     * String.
     */
    public static Symbol gensym(String name) {
	String s;
	do {
	    gencounter += 1;
	    if (gencounter == Integer.MAX_VALUE) {
		throw new Error("Symbol.gensym ran out of integers!");
	    }
	    s = name + Integer.toString(gencounter);
	} while (symbolTable.get(s) != null);
	return intern(s);
    }
    /**
     * Counter used by gensym().
     */
    static int gencounter = 0;
    /**
     * Return a new uniquely-named Symbol whose name starts with the name
     * of the given Symbol.
     */
    public static Symbol gensym(Symbol other) {
	return gensym(other.name);
    }
    /**
     * Try to unify this Symbol with another Unifiable.
     * A Symbol unifies only with itself.
     */
    public boolean unifyWith(Unifiable other, BindingSet bindings) {
	Debug.debug("Symbol.unifyWith: this=" + this + ", other=" + other);
	if (other instanceof Variable) {
	    // Symbol might unify with Variable: go see
	    Debug.debug("Symbol.unifyWith: calling Variable.unifyWith");
	    return other.unifyWith(this, bindings);
	} else if (!(other instanceof Symbol)) {
	    // Otherwise Symbol only unifies with another Symbol
	    Debug.debug("Symbol.unifyWith: other not Symbol or Variable: returning false");
	    return false;
	}
	// Symbol only unifies with itself (with no additional bindings)
	Debug.debug("Symbol.unifyWith: other is Symbol: returning " + (this == other));
	return (this == other);
    }	    
    /**
     * Returns this Symbol itself, since applying bindings to a Symbol
     * has no effect and symbols don't need to be copied.
     */
    public Unifiable copyWithBindings(BindingSet bindings) {
	return this;
    }
}
