/*
 * Variable.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 24 Apr 2002
 * Time-stamp: <Wed May 14 12:24:53 EDT 2003 ferguson>
 */

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Set;

/**
 * Variables can be constrained with a list of possible values.
 * Different Variables can have the same name (a Symbol).
 */
public class Variable implements Unifiable {
    /**
     * The name of this Variable.
     */
    Symbol name;
    /**
     * The possible values of this Variable (if any, otherwise unconstrained).
     */
    VariableValueSet values;
    /**
     * Create an unconstrained Variable with given name.
     */
    public Variable(Symbol name) {
	this(name, null);
    }
    /**
     * Create a new Variable with then given name and value restriction.
     */
    public Variable(Symbol name, VariableValueSet values) {
	this.name = name;
	this.values = values;
    }
    /**
     * Returns the Set of possible values for this Variable.
     */
    public Set getValues() {
	return values;
    }
    /**
     * Returns the printable representation of this Variable as a
     * String.
     * Variables print with a leading question mark, followed by their
     * name, then their list of constrained values, if any.
     */
    public String toString() {
	if (values == null || values.isEmpty()) {
	    return "?" + name.toString();
	} else {
	    StringBuffer buf = new StringBuffer();
	    Iterator vals = values.iterator();
	    try {
		buf.append("(?");
		buf.append(name);
		buf.append(" ");
		while (vals.hasNext()) {
		    buf.append(vals.next());
		    if (vals.hasNext()) {
			buf.append(" ");
		    }
		}
		buf.append(")");
	    } catch (NoSuchElementException ex) {
	    }
	    return buf.toString();
	}
    }
    /**
     * Try to unify this Variable with another Unifiable.
     * A Variable unifies as follows:
     * <UL>
     * <LI>An unconstrained Variable unifies with a Symbol by binding
     * the Variable to the Symbol.
     * <LI>A constrained Variable unifies with a Symbol if the Symbol
     * is one of the acceptable values, by binding the Variable to the Symbol.
     * <LI>A Variable unifies with an Expression by binding the Variable to the
     * Expression <B>IF</B> the Variable does not occur in the Expression.
     * <LI>A Variable unifies with another Variable as follows:
     * <UL>
     * <LI>If either Variable is unconstrained, it is bound to the other
     * (if both are unconstrained, it doesn't matter which is chosen).
     * other.
     * <LI>If both Variables are constrained, consider the intersection of the
     * constraints:
     * <UL>
     * <LI>If the intersection is empty, unification fails.
     * <LI>If the intersection contains just one element, bind both Variables
     * to that element.
     * <LI>Otherwise bind both Variables to a new Variable constrained
     * to the intersection values.
     * </UL>
     * </UL>
     * </UL>
     */
    public boolean unifyWith(Unifiable other, BindingSet bindings) {
	Debug.debug("Variable.unifyWith: this=" + this + ", other=" + other);
	// Check if we are already bound
	Unifiable myvalue = (Unifiable)bindings.getBinding(this);
	if (myvalue != null) {
	    // Yes we are => try to unify with that instead
	    Debug.debug("Variable.unifyWith: already bound to " + myvalue);
	    return myvalue.unifyWith(other, bindings);
	}
	// Otherwise we are an unbound Variable
	if (other instanceof Symbol) {
	    Debug.debug("Variable.unifyWith: other is Symbol");
	    // Variable unifies with Symbol if...
	    if (values == null) {
		// Unconstrained variable unifies with any Symbol
		Debug.debug("Variable.unifyWith: unconstrained: binding to Symbol " + other);
		bindings.setBinding(this, other);
		return true;
	    } else if (values.contains(other)) {
		// Constrained var unifies with Symbol if it's an acceptable value
		Debug.debug("Variable.unifyWith: constraint ok: binding to Symbol " + other);
		bindings.setBinding(this, other);
		return true;
	    } else {
		Debug.debug("Variable.unifyWith: constraint failed for Symbol " + other);
		return false;
	    }
	} else if (other instanceof Expression) {
	    Debug.debug("Variable.unifyWith: other is Expression");
	    // Variable unifies with Expression unless it occurs therein
	    if (occursIn((Expression)other)) {
		Debug.debug("Variable.unifyWith: occurs check failed");
		return false;
	    } else {
		Debug.debug("Variable.unifyWith: occurs check ok: binding to Expression " + other);
		bindings.setBinding(this, other);
		return true;
	    }
	} else if (other instanceof Variable) {
	    Debug.debug("Variable.unifyWith: other is Variable");
	    // Variable unifying with Variable has several cases
	    Variable otherv = (Variable)other;
	    if (values == null) {
		// If we are unconstrained, then bind us to them
		Debug.debug("Variable.unifyWith: we are unconstrained: binding us to other " + other);
		bindings.setBinding(this, other);
		return true;
	    } else if (otherv.values == null) {
		// If other var is unconstrained, then bind it to us
		Debug.debug("Variable.unifyWith: other is unconstrained: binding other to us " + this);
		bindings.setBinding(otherv, this);
		return true;
	    } else {
		Debug.debug("Variable.unifyWith: both Variables are constrained");
		// Otherwise we are both constrained: compute intersection
		VariableValueSet intset = values.intersectWith(otherv.values);
		if (intset.isEmpty()) {
		    // Intersection empty => fail
		    Debug.debug("Variable.unifyWith: intersection empty: fail");
		    return false;
		} else if (intset.size() == 1) {
		    // Intersection leaves one elt: bind both vars to it
		    Object[] elts = intset.toArray();
		    Debug.debug("Variable.unifyWith: intersection unit: " + elts[0]);
		    bindings.setBinding(this, elts[0]);
		    bindings.setBinding(otherv, elts[0]);
		    return true;
		} else {
		    // Otherwise we have a (possibly more) constrained var
		    Variable newvar = new Variable(name, intset);
		    Debug.debug("Variable.unifyWith: new var: " + newvar);
		    bindings.setBinding(this, newvar);
		    bindings.setBinding(otherv, newvar);
		    return true;
		}
	    }
	} else {
	    // What else could there be?
	    Debug.debug("Variable.unifyWith: failed for unknown reason!");
	    return false;
	}
    }	    
    /**
     * Returns true if this Variable occurs in the given Expression.
     */ 
    boolean occursIn(Expression expr) {
	Iterator elts = expr.iterator();
	try {
	    while (elts.hasNext()) {
		Object elt = elts.next();
		// I think == is ok; don't need equals() (Lisp would be EQ)
		if (elt == this) {
		    return true;
		} else if (elt instanceof Expression) {
		    if (occursIn((Expression)elt)) {
			return true;
		    }
		}
	    }
	} catch (NoSuchElementException ex) {
	}
	return false;
    }
    /**
     * Returns a new Variable which is a copy of this one with the
     * given bindings applied. That is, if this Variable has a binding
     * in bindings, then we return a copy of that, otherwise we return
     * a copy of the Variable with the same name.
     */
    public Unifiable copyWithBindings(BindingSet bindings) {
	Debug.debug("Variable.copy: this=" + this);
	// Do we have a binding?
	Unifiable myvalue = (Unifiable)bindings.getBinding(this);
	if (myvalue != null) {
	    // Yes, use that instead
	    Debug.debug("Variable.copy: returning binding " + myvalue);
	    return myvalue;
	} else {
	    // No, create a new Variable like me only different
	    // Note: Don't have to copy the values since they are never changed
	    Variable newvar = new Variable(name, values);
	    Debug.debug("Variable.copy: copied var " + newvar);
	    // Add to bindings so subsequent instances of me get it
	    bindings.setBinding(this, newvar);
	    // And return the new Variable
	    return newvar;
	}
    }
}
